/* 
Version vom 19.09.2012
 Realisiert mit Tiny44
 
Quarz mit 8MHz, 
 
 Layout:

           +3.3V VCC 1  14 GND
              Q1 PB0 2  13 PA0 ADC0 Ubatt/7
              Q2 PB1 3  12 PA1 Status-LED gruen
           Reset PB3 4  11 PA2 Status-LED rot
     TxBluetooth PB2 5  10 PA3
   IR-Empfaenger PA7 6  9  PA4 SCK
            MOSI PA6 7  8  PA5 MISO       


Bisher:

                      Reset  1    8  +5V
         Ubatt/3 (ADC3) PB3  2     7  PB2  Status-LED Gruen
         IR-Empfaenger  PB4  3     6  PB1  Status-LED Rot
                        GND  4     5  PB0  TX nach Bluetooth

 Input: 
   PA0  runtergeteilte Batteriespannung (vor dem 3,3V Spannungsregler). 
        Spannungsteiler: 33K/5K6  (1:7)
	      Kann also bis 7,7V messen (1.1V Uref). 
   PA7  IR-Empfnger Vishay  TSOP31238
 Output:
   PB2  Ausgang zum Bluetooth Modul
   PA1  Ausgang zur Status-LED gruen
   PA2  Ausgang zur Status-LED rot  
*/

/*
  Empfngt Bursts von 12 Pulsen mit 38KHz. Dazwischen 16 Zyklen Pause.

   Ein Durchlauf dauert also 28 Pulse. Bei 28.3s Zykluszeit sind das 792.4s,
   entsprechend 1262Hz. Damit ist eine Messung mit Millisekundenauflsung mglich.
   
   Wenn IR-Signale eintreffen ist die Status-LED grn, rot wenn nicht. 
   Die LED ist aus wenn die Schaltung aus ist (oder Ubatt zu gering fr Prozessorbetrieb)
   Statt Duo-LED knnen auch 2 antiparallel geschaltete LEDs verwendet werden.

*/

/*
Kompaktformat: 12 Bytes

Um mit Terminal debugbar zu bleiben werden nur ASCII-Zeichen verwendet.


Byte  Inhalt            Wertebereich
 1    Start-Kennung     10
 2    Life Indikator      64+ (32+16)   Bit 4,5
 2    Seqenznummer        64+ (8+4+2+1) Bits 0..3    
 3    T1 MSB               32+T1/64  Flugzeit in Millisekunden
 4    T1 LSB               32+T1%64 
 5    T2 MSB               32+T2/64 
 6    T2 LSB               32+T2%64 
 7    T3 MSB               32+T3/64 
 8    T3 LSB               32+T3%64 
 9    T4 MSB               32+T4/64 
 10   T4 LSB               32+T4%64 
 11   Ubatt                50-76       in 1/10V
 12   Ende-Kennung      13

*/

/* Optimizer auf -O1 */

#define F_CPU 8000000UL	/* CPU clock in Hertz */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <string.h>
//#include <stdio.h>
//#include <stdint.h>

#define LED_ROT {PORTA&=~BIT1; PORTA|=BIT2;}
#define LED_GRUEN {PORTA&=~BIT2; PORTA|=BIT1;}
#define LED_AUS {PORTA&=~BIT2; PORTA&=~BIT1;}
#define TX_LOW  {PORTB&=~BIT2;}
#define TX_HIGH {PORTB|=BIT2;}
#define LICHTSCHRANKE_UNTERBROCHEN (PINA & BIT7)  

#define TIMER_EIN TIMSK1=(1<<OCIE1A); TIFR1=(1<<OCF1A)
#define TIMER_AUS TIMSK1=0





#define U8 unsigned char
#define U16 unsigned int
#define U32 unsigned long

#define BIT0 1
#define BIT1 2
#define BIT2 4
#define BIT3 8
#define BIT4 16
#define BIT5 32
#define BIT6 64
#define BIT7 128



typedef struct{
  U8 signal;  
  U16 flugzeit;
  U8 neueMessung;  

  U8 senden;
  U16 t1,t2,t3,t4;
}Tglob;

Tglob volatile g;








//------------------------------------------------------------------------------------------------------------ 
ISR(TIM0_COMPA_vect){ // 19230Hz (52s)
//------------------------------------------------------------------------------------------------------------ 
static U8 puls=0;
static U8 tiefpass=0;
static U8 filter=0;
static U16 sperrzeit=0;
static U8 licht=0;
static U8 bisher=9;

static U8 TxZyklus=0;
static U8 status=0;
static U8 bitmask=1;
static U8 b;
static U32 us; //zhlt Mikrosekunden

  //Serielle Schnittstelle in jedem 2. Zyklus bedienen (104s->9615Hz)
  TxZyklus ^=1;
  if (TxZyklus){  
    //Sendet das in "g.senden" enthaltene Byte mit 9600 Baud und setzt g.senden auf 0.
    //'\0'kann daher nicht gesendet werden!
    switch(status){
      case 0:  //warten auf Auftrag
        if (g.senden){
          b=g.senden;
          g.senden=0;
          TX_LOW; //Startbit
          bitmask=1;
          status=1;
        }  
        break;

      case 1: //Data bits
        if (b & bitmask){
          TX_HIGH;
        }else{
          TX_LOW;
        }
        bitmask<<=1;
        if  (0==bitmask) status=2;   
        break;

      case 2: //Stoppbit 1
        TX_HIGH; 
        status=3;
        break;

      case 3: //Stoppbit 2
        status=0;
        break;
    }
  }
  
  

  //Bei nicht unterbrochener Lichtschranke kommen Pulse mit 320s Low und Pausen mit ca. 450s High, 
  //Pulserkennung: lnger als 260s low
  if (LICHTSCHRANKE_UNTERBROCHEN){ //High
    if (tiefpass>0) tiefpass--; else puls=0;    //AUS nach 16*52=832s
  }else{                          //Low
    if (tiefpass<15) tiefpass+=3; else puls=1;  //EIN nach 5*52=260s
  }

  //Mehr als 3 Pulse/fehlende Pulse hintereinander dienen zur Erkennung
  if (puls){ 
    if (filter<45) filter++; 
    else{
      licht=1;  //Filter ber 3 Bursts:  792s/52s=15,2 Zyklen.  45=3*15
      LED_GRUEN;
    }

  }else{
    if (filter>0) filter--; 
    else{
      licht=0;
      LED_ROT;
    } 
  }

  
  us+=52; //Stoppuhr 52 Mikrosekunden weiterdrehen 
  if (sperrzeit!=0)  sperrzeit--;

  if ((licht !=bisher) && (0==sperrzeit)){  //Flanke erkannt
    bisher=licht;
    if (licht){ //Absprung
      sperrzeit=  1923; //100ms Sperrzeit: 1923*52s=100.000s
      us=0;  //neue Zeitmessung 
      //LED_GRUEN;
    }else{ //Landung
      g.flugzeit = us/1000; 
      g.neueMessung++;        
      sperrzeit=  1000; // 52ms Sperrzeit: 
      //LED_ROT;
    }
  }
}



//------------------------------------------------------------------------------------------------------------ 
void sendbyte(U8 b){
//------------------------------------------------------------------------------------------------------------ 
  while(g.senden)
  ;
  g.senden=b;
}

//------------------------------------------------------------------------------------------------------------ 
void sendLine(U8 volatile *txt){
//------------------------------------------------------------------------------------------------------------ 
  while(*txt){
    sendbyte(*txt++);
  }
}



//------------------------------------------------------------------------------------------------------------ 
int main(void){
//------------------------------------------------------------------------------------------------------------ 
static U8 txbuffer[15];

U8 Ubatt;
U16 tmp;
U8 sn=0; //Sequenznummer der Sprnge (fortlaufende einstellige Zahl)
U8 z=0;
U8 *p;
U8 loop=0;



  cli();
  DDRA = BIT1|BIT2;
  DDRB =      BIT2;
  LED_ROT;


  ADMUX=(1<<REFS1) ; //Uref=1.1, input on ADC0
  ADCSRA=(1<<ADEN) | (1<<ADPS2) | (1<<ADPS1) | (1<<ADPS0); // 8MHz/128 gibt 62,5KHz
  ADCSRB=(1<<ADLAR);


  ACSR=(1<<ACD); //Analogkomparator aus
  PRR=(1<<PRTIM1)|(1<<PRUSI); //Timer1 und USI aus



  //Timer 0: 19230Hz Interrupts (um 9600Baud "von Hand" zu machen)
  TCCR0A=(1<<WGM01);  //CTC
  TCCR0B=(1<<CS01);   //Teiler 8 (1MHz)

  //Es wird nicht durch 104 geteilt (9615Hz) sondern der Timer luft doppelt so schnell
  //Damit kann ich die Lichtschrankenpulse besser gefiltert auswerten (52s Raster)
  OCR0A = 51; // Teiler / 52 - > 19230Hz
  OCR0B = 0; 

  GTCCR = 0;
  TIMSK0=(1<<OCIE0A);

  memset((U8 *)&g, sizeof(Tglob),0);

  g.t1=1111;
  g.t2=2222;
  g.t3=3333;
  g.t4=4444;

  sei(); //Interrupts scharf.

#if 0
  //Testpattern senden
  for (n=0; n<10; n++){
    g.senden=10;
    _delay_ms(100);
  }
#endif
  //...................................................................................... 
  while(1){
    //Batteriespannung messen
    ADCSRA|= (1<<ADSC) ; //Starte Messung

    
    _delay_ms(100);  //Begrenze auf 10 Datenpakete je Sekunde
    Ubatt = ADCH;
    if (g.neueMessung){
      g.neueMessung=0;  
      if ((g.flugzeit >= 300) && (g.flugzeit < 4000)){
        sn++;   //Sequenznummer
        sn &=15;
        g.t4=g.t3;
        g.t3=g.t2;
        g.t2=g.t1;
        g.t1=g.flugzeit;
      }
    }else{  
      if (loop++ < 10) continue;
    }
    loop=0;

    switch(z){ //Life-Indicator
      case 0: z=16; break;
      case 16: z=32; break;
      case 32: z=48; break;
      default:  z=0;break;
    }

    p = txbuffer;
    *p++=10;            //1
    *p++=64 | z | sn;   //2
    *p++=32 + g.t1/64;  //3
    *p++=32 + g.t1%64;
    *p++=32 + g.t2/64;  //5
    *p++=32 + g.t2%64;
    *p++=32 + g.t3/64;  //7
    *p++=32 + g.t3%64;
    *p++=32 + g.t4/64;  //9
    *p++=32 + g.t4%64;


    tmp = Ubatt;  //256 entspricht 7.7 V
    tmp*=3;       //768 entspricht 7.7V - also ungefhr Faktor 100
    tmp/=10;      //Umrechnung in 1/10V
    if (tmp<32) tmp=32;
    if (tmp>77) tmp=77;
    *p++=tmp;          //11
    *p++=13;            //12 
    *p=0;               //13         
    sendLine(txbuffer);
  }
}
//EOF
